# SPDX-FileCopyrightText: Portions Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

PROBE_RGB = "e5a70e"
DETECTOR_RGB = "2c79da"
GENERATOR_RGB = "c061cb"

EMOJI_SCALE_FACE = "😭🙁😐🙂🤩"
EMOJI_SCALE_COLOUR = "🔻🔶🟡🟩💙"
EMOJI_SCALE_COLOUR_SQUARE = "🟥🟧🟨🟩🟦"
EMOJI_SCALE_ALT = "🚫⚠️👌✅🚀"
