<!DOCTYPE html>
<html lang="en">

<head>
<meta name="viewport" content="width=device-width, initial-scale=1">
<meta charset="UTF-8" />
<style>
body {font-family: sans-serif}
:root{
  --defcon1: #f94144;
  --defcon2: #f8961e;
  --defcon3: #ccc;
  --defcon4: #eee;
  --defcon5: #f7f7ff;
}
.defcon1 {background-color: var(--defcon1); text-color: #000}
.defcon2 {background-color: var(--defcon2); text-color: #000}
.defcon3 {background-color: var(--defcon3); text-color: #000}
.defcon4 {background-color: var(--defcon4); text-color: #000}
.defcon5 {background-color: var(--defcon5); text-color: #000}
.probe {padding-left: 40pt}
.detector {padding-left: 65pt}
.zscore {
  padding-top: 6pt; 
  padding-bottom: 6pt; 
  margin-left: 60pt;
  border: 1pt solid #ccc;
}
.zscore b {
  padding: 6pt 10pt 7pt 10pt; 
  margin: 0
}
h2 {padding-left: 20pt}
h3 {padding-left: 40pt}
h4 {padding-left: 60pt}
h2,h3,h4 {padding-top: 10px; padding-bottom: 10px}

/* Style the buttons that are used to open and close the accordion panel */
.accordion {
//  background-color: #eee;
  color: #444;
  cursor: pointer;
  padding: 18px;
  width: 100%;
  text-align: left;
  border: none;
  outline: none;
  transition: 0.4s;
  margin: 1pt;
}

/* Add a background color to the button if it is clicked on (add the .active class with JS), and when you move the mouse over it (hover) */
.active, .accordion:hover {
  background-color: #ccf;
}

/* Style the accordion panel. Note: hidden by default */
.panel {
  padding: 0 18px;
  background-color: white;
  display: none;
  overflow: hidden;
}
</style>
<title>garak report: {{reportfile}}</title>
<meta name="description" content="" />
</head>

<body>

<h1>garak run: {{reportfile}}</h1>
<button class="accordion">⚙️ view config</button>
<div style="border:solid black 1px; padding: 5px; margin: 5px" class="panel">
<h2>config details</h2>
<pre>
filename: {{reportfile}}

garak version: {{garak_version}}

generator: {{model_type}}.{{model_name}}

probe spec: {{probespec}}

run started at: {{start_time}}

run config: {{setup}}

{% for payload in payloads %}
payload: {{payload}}

{% endfor %}
</pre>
</div>

{%if model_name %}
<h2>results: {{model_type}} / {{model_name}}</h2>
{%else%}
<h2>results:</h2>
{%endif%}